;;; This file is part of guile-with-graphics, a graphic library for Guile

;;; Copyright (C) 2009  Luca Saiu
;;; Updated in 2010 by Luca Saiu

;;; This program is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(use-modules (srfi srfi-9)) ;; this is needed for records
(use-modules (ice-9 format))
;;; The game field size, in pixels:
(define field-width  800)
(define field-height 600)

;;; The "border" size, i.e. the number of pixels within the field which are
;;; impossible to reach for players:
(define field-border 20)
(define minimum-x-in-the-field field-border)
(define minimum-y-in-the-field field-border)
(define maximum-x-in-the-field (- field-width field-border))
(define maximum-y-in-the-field (- field-height field-border))
(define (in-field-x? x)
  (and (<= minimum-x-in-the-field x)
       (<= x maximum-x-in-the-field)))
(define (in-field-y? y)
  (and (<= minimum-y-in-the-field y)
       (<= y maximum-y-in-the-field)))

;;; Player:
(define player-radius 12) ;12)
(define 2player-radius (* player-radius 2))

;;; How many updates per second to compute and to send to clients (by default).
;;; Notice that about ten decisions per seconds are more enough to play well;
;;; it's still faster than the human reaction time, and with a small frequency
;;; we don't stress the processor too much.
(define default-server-update-recomputation-frequency 62)
(define default-server-update-sending-frequency 30)
(define server-update-sending-frequency-for-client 10);60)
(define server-update-sending-frequency-for-ai-client 15)

;;; We represent the current state of each pill as a record:
(define-record-type pill
  (make-pill-by-specifying-all-fields x y velocity-x velocity-y)
  pill?
  (x                 pill-x                 pill-set-x!)
  (y                 pill-y                 pill-set-y!)
  (velocity-x        pill-velocity-x        pill-set-velocity-x!)
  (velocity-y        pill-velocity-y        pill-set-velocity-y!))

(define (make-pill)
  (make-pill-by-specifying-all-fields (random-x-in-the-field)
                                      (random-y-in-the-field)
                                      (random-from-to (- maximum-pill-component-velocity-in-modulo)
                                                      maximum-pill-component-velocity-in-modulo)
                                      (random-from-to (- maximum-pill-component-velocity-in-modulo)
                                                      maximum-pill-component-velocity-in-modulo)))

;;; Pills:
(define pills-no 20);30)
(define pill-radius 5)
(define pill-eating-radius (max (* player-radius 3/2)
                                (* pill-radius 3/2)))

;;; Maximum translation velocity in modulo for pills, for each component (x and y),
;;; in pixels per second:
(define maximum-pill-component-velocity-in-modulo
  50)

;;; Rotation velocity, in radiants per second:
(define rotation-velocity 9)

;;; Translation velocity, in pixels per second:
(define translation-velocity 200);300)

;;; Compute random reachable coordinates within the field:
(define (random-x-in-the-field)
  (random-from-to field-border maximum-x-in-the-field))
(define (random-y-in-the-field)
  (random-from-to field-border maximum-y-in-the-field))

;;; Saturate coordinates (meaning: return them unchanged if they are within
;;; the game field, otherwise return the appropriate border value):
(define (saturate-x x)
  (saturate x field-border maximum-x-in-the-field))
(define (saturate-y y)
  (saturate y field-border maximum-y-in-the-field))

;;; Compute the Euclidean distance from the given position to the given pill:
(define (distance-from-pill x y pill)
  (distance (pill-x pill) (pill-y pill) x y))

;;; Is the given pill within the eating range from the given position?
(define (eaten? x y pill)
  ;; Notice that a pill is represented as the cons of its coordinates:
  (<= (distance-from-pill x y pill)
      pill-eating-radius))

;;; Return the angle which must be added to the given angle of an object in the
;;; given position so that the object in the given position is oriented towards
;;; the given point:
(define (angle-to-target x y angle target-x target-y)
  (let* ((delta-x (- target-x x))
         (delta-y (- y target-y))
         (delta-x (if (zero? delta-x) 0.0001 delta-x)))
    (normalize-angle (- (if (> delta-x 0)
                            (atan (/ delta-y delta-x))
                            (+ pi (atan (/ delta-y delta-x))))
                        angle))))

(initialize-graphics "Interactive game standalone demo" field-width field-height)

;;; I pre-define some useful colors. Of course, always remember that color
;;; coordinates are numbers (floats like 0.7, but also also *rationals* 
;;; like 1/3 are accepted in Scheme) between 0 and 1:
(define black   (make-color 0   0   0))
(define white   (make-color 1/2 1/2 1/2))
(define red     (make-color 1/2 0   0))
(define green   (make-color 0   1/2 0))
(define blue    (make-color 0   0   1/2))
(define cyan    (make-color 0   1/2 1/2))
(define magenta (make-color 1/2 0   1/2))
(define yellow  (make-color 1/2 1/2 0))
(define gray    (make-color 1/2 1/2 1/2))

;;; Return a new color obtained by multiplying each component of the given
;;; color by the given factor, and saturating:
(define (color-* factor color)
  (let ((red   (color->red color))
        (green (color->green color))
        (blue  (color->blue color)))
    (make-color (saturate (* factor red) 0 1)
                (saturate (* factor green) 0 1)
                (saturate (* factor blue) 0 1))))

;;; Given a color, return a ligher version of it. For example (light green)
;;; makes a lighter green:
(define (light color)
  (color-* 3/2 color))

;;; Given a color, return a darker version of it. For example (dark green)
;;; makes a darker green:
(define (dark color)
  (color-* 2/3 color))

;; Return a new random color:
(define (random-color)
  (make-color (random 1.0) (random 1.0) (random 1.0)))

;; This is the most general version of for, but it's not always
;; comfortable to call. I use this to define three friendlier
;; versions:
(define-macro (for-step variable from to step . body)
  (let ((to-name (gensym))
        (from-name (gensym))
        (step-name (gensym))
        (lower-limit-name (gensym))
        (upper-limit-name (gensym)))
    `(let* ((,from-name ,from)
            (,to-name ,to)
            (,variable ,from-name)
            (,lower-limit-name (min ,from-name ,to-name))
            (,upper-limit-name (max ,from-name ,to-name))
            (,step-name ,step))
       (while (and (>= ,variable ,lower-limit-name)
                   (<= ,variable ,upper-limit-name))
              ,@(append body
                        `((set! ,variable (+ ,variable ,step-name))))))))

;; Evaluate the body a number of times equal to how-many-times.
;; No iteration variable is visible in the body,
;;
;; Example: (n-times 5 (display "This will be printed five times\n")
(define-macro (n-times how-many-times . body)
  (let ((variable (gensym)))
    `(for ,variable 1 ,how-many-times ,@body)))

;; The classic for loop, with the variable incremented at each iteration.
;;
;; for example (for i 1 5
;;               (display "i is now ")
;;               (display i)
;;               (newline))
;; is equivalent to the C loop:
;; { 
;;   int i;
;;   for(i = 1; i <= 5; i++){
;;     printf("i is now ");
;;     printf("%i", i);
;;     printf("\n");
;;   } // for
;; }
(define-macro (for variable from to . body)
  `(for-step ,variable ,from ,to 1 ,@body))

;; Just like for, but here the variable is *de*cremented at each iteration:
(define-macro (down-for variable from to . body)
  `(for-step ,variable ,from ,to -1 ,@body))

;;; Return any number in the interval [minimum, maximum], with an uniform
;;; probability distribution:
(define (random-from-to minimum maximum)
  (if (and (integer? minimum)
           (integer? maximum))
      (+ (random (- maximum minimum -1))
         minimum)
      (+ (random (exact->inexact (- maximum minimum)))
         minimum)))

;; ;;; Try this:
;; (define (rectangles-demo)
;;   (let ((x0 (random-from-to -10 (+ (window-width) 10)))
;;         (y0 (random-from-to -10 (+ (window-height) 10)))
;;         (width (random-from-to 10 (/ (window-width) 2)))
;;         (height (random-from-to 10 (/ (window-height) 2)))
;;         (color (random-color)))
;;     (handle-quit-event-only)
;;     (lock)
;;     (draw-rectangle x0 y0 width height color)
;;     (unlock)
;;     (refresh-rectangle x0 y0 width height)
;;     (rectangles-demo)))

;; ;;; Try this:
;; (define (rectangle-borders-demo)
;;   (let ((x0 (random-from-to -10 (+ (window-width) 10)))
;;         (y0 (random-from-to -10 (+ (window-height) 10)))
;;         (width (random-from-to 10 (/ (window-width) 2)))
;;         (height (random-from-to 10 (/ (window-height) 2)))
;;         (color (random-color)))
;;     (handle-quit-event-only)
;;     (lock)
;;     (draw-rectangle-border x0 y0 width height color)
;;     (unlock)
;;     (refresh-rectangle x0 y0 width height)
;;     (rectangle-borders-demo)))

;; ;;; Try this:
;; (define (lines-demo)
;;   (let ((x1 (random (window-width)))
;;         (y1 (random (window-height)))
;;         (x2 (random (window-width)))
;;         (y2 (random (window-height)))
;;         (color (random-color)))
;;     (handle-quit-event-only)
;;     (lock)
;;     (draw-line x1 y1 x2 y2 color)
;;     (unlock)
;;     (refresh-rectangle x1 y1 (- x2 x1) (- y2 y1))
;;     (lines-demo)))

;; (define (block-grid-demo)
;;   (let* ((block-width 60)
;;          (block-height 20)
;;          (width (window-width))
;;          (height (window-height)))
;;     (handle-quit-event-only)
;;     (for x 0 (/ width block-width)
;;          (for y 0 (/ height block-height)
;;               (lock)
;;               (draw-block (* x block-width) (* y block-height)
;;                           (- block-width 4) (- block-height 4)
;;                           ;;(dark (dark red))
;;                           (random-color)
;;                           3)
;;               (unlock)))
;;     (refresh)
;;     (block-grid-demo)))

;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; (define block-width 80)
;; (define block-height 25)
;; (define block-color (dark red))
;; (define block-x (/ (window-width) 2))
;; (define block-y (/ (window-height) 2))
;; (define delta-x 0)
;; (define delta-y 0)
;; (define changing-color #f)
;; (define time-at-startup (time))
;; (define tick-length 30)
;; (define speed (* 0.2 tick-length))

;; (define (handle-event event)
;;   (cond ((or (equal? event '(quit))
;;              (equal? event '(key-down #\q))
;;              (equal? event '(key-down escape)))
;;          (exit 0))
;;         ((equal? event '(key-down left))
;;          (set! delta-x (- speed)))
;;         ((equal? event '(key-up left))
;;          (set! delta-x 0))
;;         ((equal? event '(key-down right))
;;          (set! delta-x speed))
;;         ((equal? event '(key-up right))
;;          (set! delta-x 0))
;;         ((equal? event '(key-down down))
;;          (set! delta-y speed))
;;         ((equal? event '(key-up down))
;;          (set! delta-y 0))
;;         ((equal? event '(key-down up))
;;          (set! delta-y (- speed)))
;;         ((equal? event '(key-up up))
;;          (set! delta-y 0))
;;         ((equal? event '(key-up space))
;;          (set! changing-color #f)) 
;;         ((equal? event '(key-down space))
;;          (set! changing-color #t))
;;         (else
;;          (format #t "Unhandled event: ~a\n" event))))

;; (define (on-tick)
;;   (set! block-x (+ block-x delta-x))
;;   (set! block-y (+ block-y delta-y))
;;   (if changing-color
;;       (set! block-color (random-color)))
;;   (draw-block block-x block-y block-width block-height block-color)
;;   (refresh)
;;   ;;(refresh-rectangle (- block-x 1) (- block-y 1) (+ block-width 2) (+ block-height 2))
;;   (draw-rectangle (- block-x 1) (- block-y 1) (+ block-width 2) (+ block-height 2)
;;                   black))


;; (clear blue)
;; (refresh)

;; ;(rectangles-demo)
;; ;(rectangle-borders-demo)
;; ;(blocks-demo)
;; ;(lines-demo)
;; ;(block-grid-demo)

;; (interactive-game tick-length handle-event on-tick)

;;; Return the n-th element of the given list:
(define (n-th n list)
  (if (zero? n)
      (car list)
      (n-th (- n 1)
            (cdr list))))

;;; Convert the given s-expression into its printer representation:
(define (s-expression->string s-expression)
  (let ((string-port (open-output-string)))
    (write s-expression string-port)
    (get-output-string string-port)))

;;; Some useful constants:
(define pi (* 2 (acos 0)))
(define -pi (- pi))
(define 2pi (* 2 pi))
(define pi/2 (/ pi 2))
(define pi/3 (/ pi 3))
(define pi/4 (/ pi 4))
(define pi/6 (/ pi 6))

;;; Given an angle in radiants, return another equivalent one in (-pi, +pi]
(define (normalize-angle angle)
  (cond ((> angle pi)
         (normalize-angle (- angle 2pi)))
        ((<= angle -pi)
         (normalize-angle (+ angle 2pi)))
        (else
         angle)))

;;; Destructively update a variable representing an angle in radiants,
;;; replacing them with the normalized value:
(define-macro (normalize-angle! variable)
  `(set! ,variable (normalize-angle ,variable)))

;;; Convert an angle from radiants to degrees. This is probably only useful for debugging:
(define (radiants->degrees angle)
  (* 180 (/ angle pi)))

;;; Return a copy of the given list without the given element (if it is in the
;;; original list):
(define (without-with-equality-predicate equality-predicate element list)
  (define (without element list accumulator)
    (cond ((null? list)
           accumulator)
          ((equality-predicate element (car list))
           (append (cdr list) accumulator))
          (else
           (without element (cdr list) (cons (car list) accumulator)))))
  (without element list '()))

;;; The "default" function simply called without uses equal? (structural comparison),
;;; and is just an alias for without-by-structure:
(define (without-by-structure element list)
  (without-with-equality-predicate equal? element list))
(define without without-by-structure)
 
;;; without-by-identity uses eq? (equality by identity)
(define (without-by-identity element list)
  (without-with-equality-predicate eq? element list))

;;; Return a copy of the first parameter without the elements in the second parameter.
;;; This is an implementation of set subtraction, where; sets are implemented as unsorted lists.
(define (subtract-lists-with-equality-function equality-function? list elements-to-remove)
  (if (null? elements-to-remove)
      list
      (subtract-lists-with-equality-function equality-function?
                                             (without-by-structure (car elements-to-remove) list)
                                             (cdr elements-to-remove))))
(define (subtract-lists-by-identity list elements-to-remove)
  (subtract-lists-with-equality-function eq? list elements-to-remove))
(define (subtract-lists-by-structure list elements-to-remove)
  (subtract-lists-with-equality-function equal? list elements-to-remove))

;;; Call the given function on the elements of the given list, in order. The results
;;; are ignored:
(define (iterate function list)
  (if (null? list)
      #t
      (begin
        (function (car list))
        (iterate function (cdr list)))))

;;; Common Lisp-style when and unless. These are one-way if conditionals, which take
;;; one or more forms as their body even without using begin. when executes its body
;;; (and returns the value of its last form) if the condition is true, and unless
;;; executes its body (and returns the value of its last form) if the condition is
;;; false:
(define-macro (when condition . body)
  `(if ,condition
       (begin
         ,@body)))
(define-macro (unless condition . body)
  `(if (not ,condition)
       (begin
         ,@body)))

;;; Useful higher-order functions working on lists (note that map and filter
;;; are predefined):
(define (for-all? predicate? list)
  (cond ((null? list)
         #t)
        ((predicate? (car list))
         (for-all? predicate? (cdr list)))
        (else
         #f)))
(define (exists? predicate? list)
  (cond ((null? list)
         #f)
        ((predicate? (car list))
         #t)
        (else
         (exists? predicate? (cdr list)))))

;;; Executes the forms in body and return the result of the last one. If any
;;; of them throws an exception, just capure it, stop executing the body and
;;; return #f:
(define-macro (ignore-failures . body)
  `(catch #t
     (lambda ()
       ,@body)
     (lambda exception
       (format #t "ignore-failures: ignored the exception ~s\n" exception)
       #f)))

(define (square x)
  (* x x))

;;; Compute the Euclidean distance between two points:
(define (distance x1 y1 x2 y2)
  (sqrt (+ (square (- x1 x2))
           (square (- y1 y2)))))

;;; Return x if it's within the given bounds (included), otherwise return the
;;; limit:
(define (saturate x minimum maximum)
  (cond ((< x minimum)
         minimum)
        ((> x maximum)
         maximum)
        (else
         x)))

;;; Set the initial random state to a value depending on the current time,
;;; so that every time we start up the program the sequence of random values
;;; is different:
(define (randomize-random-state!)
  (let ((time (gettimeofday)))
    (set! *random-state*
          (seed->random-state (+ (car time)
                                 (cdr time))))))

;;; This is very useful for debugging. The given expression is evaluated and printed
;;; on the output in a clearly readable format. For example (log (+ 1 2)) prints
;;; "(+ 1 2) ==> 3" on the output.
(define-macro (log expression)
  `(format #t "~s ==> ~s\n" ',expression ,expression))

;;; Call the given procedure in a loop, at most maximum-frequency times per second
;;; -- or less frequently if the machine is not fast enough. Exit the loop if the
;;; function returns #f.
;;; The called procedure is given the time elapsed since the last call in seconds
;;; as its only parameter.
(define (call-with-maximum-frequency-till-#f maximum-frequency ; maximum-frequency is in Hz
                                             procedure)
  (let* ((minimum-period (/ 1000 maximum-frequency)) ; in milliseconds
         (minimum-period-in-seconds (/ minimum-period 1000))
         (time-before-last-period (milliseconds-since-startup))
         (should-we-go-on #t))
    (while should-we-go-on
      (let* ((time-before-this-period (milliseconds-since-startup))
             (delta-t (- time-before-this-period
                         time-before-last-period))
             (milliseconds-to-sleep (- minimum-period delta-t)))
        (if (> milliseconds-to-sleep 0)
            (begin
              (usleep (inexact->exact (floor (* milliseconds-to-sleep 1000))))
              (set! time-before-last-period (+ time-before-this-period
                                               milliseconds-to-sleep))
              (set! should-we-go-on
                    (procedure minimum-period-in-seconds)))
            (begin
              ;;(usleep 0) ; this should at least yield to another process
              (set! time-before-last-period time-before-this-period)
              (set! should-we-go-on
                    (procedure (/ delta-t 1000)))))))))

;;; Call the given procedure in an infinite loop, at most maximum-frequency times
;;; per second -- or less frequently if the machine is not fast enough.
;;; The called procedure is given the time elapsed since the last call in seconds
;;; as its only parameter.
;;;
;;; If the given procedure returns a number, that is interpreted as the new
;;; frequency to be used from that point on.
(define (call-with-maximum-frequency maximum-frequency ; maximum-frequency is in Hz
                                     procedure)
  (let ((current-maximum-frequency maximum-frequency))
    (while #t
      (call-with-maximum-frequency-till-#f
          current-maximum-frequency
          (lambda (time-since-last-call)
            (let ((new-frequency-or-nonnumber (procedure time-since-last-call)))
              (if (number? new-frequency-or-nonnumber)
                (begin
                  (set! current-maximum-frequency new-frequency-or-nonnumber)
                  #f)     ; stop; a new loop will begin with the new frequency
                #t))))))) ; loop again

;; (define original-display
;;   (if (defined? 'original-display)
;;      original-display
;;      display))
;; (define (display . arguments)
;;   (original-display (get-internal-real-time))
;;   (original-display " ")
;;   (apply original-display arguments)
;;   (flush-all-ports)
;;   ;(usleep 1000)
;;   )

(define (fatal . arguments)
  (format #t "Fatal error: ~s\nExiting...\n" arguments)
  (kill-all-threads))

(define original-format
  (if (defined? 'original-format)
      original-format
      format))
(define stupid-mutex (make-mutex))
(define (format . stuff)
  (with-mutex stupid-mutex
    (apply original-format stuff)))



;;; Draw a *non-filled* ractangle. The interface is compatible with
;;; draw-rectangle:
(define (draw-rectangle-border x0 y0 width height color)
  (draw-line x0 y0 (+ x0 width -1) y0 color)
  (draw-line x0 y0 x0 (+ y0 height -1) color)
  (draw-line x0 (+ y0 height -1) (+ x0 width -1) (+ y0 height -1) color)
  (draw-line (+ x0 width -1) y0 (+ x0 width -1) (+ y0 height -1) color))

;;; Draw a block with lights and shadows, suitable for Tetris and Breakout;
;;; notice that there is an optional parameter (which is seen by the procedure
;;; as an empty list when not given, and a non-empty list when given):
(define (draw-block x0 y0 width height color . border-size-or-null)
  ;; If the optional parameter was supplied then bind border-size to the
  ;; value given by the user, otherwise use a default:
  (let ((border-size (if (null? border-size-or-null)
                         3
                         (car border-size-or-null))))
    ;; Draw nothing if the block is too small:
    (if (or (<= width (* 2 border-size))
            (<= height (* 2 border-size)))
        #t ; well, we have to return something, so we return #t
        ;; Ok, the block is not too small: draw it:
        (let ((lighter-color (light color)) ; color for the lightened part
              (darker-color (dark color)))  ; color for the shadowed part
          ;; Draw the shadowed part on the East and South sides:
          (draw-rectangle x0 (+ y0 height (- border-size)) width border-size darker-color)
          (draw-rectangle (+ x0 width (- border-size)) y0 border-size height darker-color)
          ;; Draw the enlightended part on the West and Nort sides:
          (draw-rectangle x0 y0 border-size height lighter-color)
          (draw-rectangle x0 y0 width border-size lighter-color)
          ;; Draw the inside part:
          (draw-rectangle (+ x0 border-size) (+ y0 border-size)
                          (- width (* 2 border-size)) (- height (* 2 border-size))
                          color))))) ;; There's no automatic refresh here!

(define (square a)
  (* a a))

;;; The absolute value should not be needed here, but I've found that some rounding
;;; errors with floats make the argument of the square root be negative, and the result
;;; of the square root complex...
;;; If you use this procedure to draw large circles the result will not look good around
;;; the leftmost and rightmost points. But it is reasonably fast, to be used for
;;; interactive applications where speed matters. The result could be made better-looking
;;; by adding another loop on the y variable.
(define (draw-circle center-x center-y radius color)
  (for x (- center-x radius) (+ center-x radius)
       (let ((delta-y (sqrt (abs (- (square radius)
                                    (square (- x center-x)))))))
         (draw-line x (- center-y delta-y) x (+ center-y delta-y) color))))
;;; See the comment above about abs. But notice that this procedure gives good results
;;; also for big circles.
(define (draw-circle-border center-x center-y radius color)
  (for x (- center-x radius) (+ center-x radius)
       (let ((delta-y (sqrt (abs (- (square radius)
                                    (square (- x center-x)))))))
         (draw-pixel x (- center-y delta-y) color)
         (draw-pixel x (+ center-y delta-y) color))))

(define nuance-step -2) ;; Make this bigger in modulo (and still negative) for slow machines
(define (draw-nuanced-circle center-x center-y maximum-radius color)
  (for-step radius maximum-radius 2 nuance-step
    (draw-circle center-x center-y radius
                 (color-* (- 1.2 (/ radius maximum-radius))
                          color))))

(define sqrt2 (sqrt 2))
;;; These can be used as an "approximation" of a circle on slow machines, for interactive
;;; applications which need speed. There is some ugly code duplication here, but the
;;; point of this code is speed, so I don't want to use a higher-order function:
(define (draw-centered-square-border center-x center-y half-diagonal color)
  (let* ((half-side (/ half-diagonal sqrt2))
         (side (* half-side 2))
         (top-left-x (- center-x half-side))
         (top-left-y (- center-y half-side)))
    (draw-rectangle-border top-left-x top-left-y side side color)))
(define (draw-centered-square center-x center-y half-diagonal color)
  (let* ((half-side (/ half-diagonal sqrt2))
         (side (* half-side 2))
         (top-left-x (- center-x half-side))
         (top-left-y (- center-y half-side)))
    (draw-rectangle top-left-x top-left-y side side color)))

;;; Let's make the console functionality always available:


;;; The user is only interested in calling console-write-line!, console-clear!
;;; and (once per tick) console-draw.  Our neat scrolling effects are
;;; automatically supported.
;;; The user can also change console-foreground-color and
;;; console-background-color with a simple set!.


;;; The user should *not* load this file directly. It's already loaded by
;;; interactive.scm.

(define console-external-border 10)
(define console-internal-border 5)
(define console-line-no 5)
(define console-inter-line-spacing 2)

;;; If the background color is #f instead of a color, then the console is drawn
;;; with no background. This variable can be safely changed with set! after
;;; initialization:
(define console-background-color #f)
;(define console-background-color (make-color 0.2 0.0 0.0))

(define console-foreground-color (make-color 1 1 1))
;;(define console-foreground-input-color (light yellow))

;;; The console state can be one of of hidden, shown, showing or hiding.
(define console-state
  'hidden)

;;; A ratio of 1 means that the console is fully shown, a ratio of 0 means that
;;; it is fully hidden. All values in [0, 1] are valid.
(define console-show-ratio 0)

;;; The absolute time (as returned by milliseconds-since-startup) when the
;;; state changed the last time:
(define console-last-state-change-time 0)

(define console-full-showing-time-after-each-change 3000) ; in milliseconds

;;; How long the console takes, in milliseconds, for showing up or hiding:
(define console-showing-time 250);;500)
(define console-hiding-time 2000)

;;; The list of lines currently shown in the console, from the oldest to the newest,
;;; as a list of strings
(define console-lines
  '())
;;'("abc" "123" "foo bar baaz quux" "boo" "this is the bottom line"))

(define (console-update-state!)
  (let ((current-time (milliseconds-since-startup)))
    (case console-state
      ((hidden)
       ;; Do nothing
       #t)
      ((shown)
       (if (>= current-time (+ console-last-state-change-time
                               console-full-showing-time-after-each-change))
           (begin
             (set! console-last-state-change-time current-time)
             (set! console-state 'hiding))))
      ((showing)
       (if (>= current-time (+ console-last-state-change-time
                               console-showing-time))
           (begin
             (set! console-last-state-change-time current-time)
             (set! console-state 'shown)
             (set! console-show-ratio 1))
           (set! console-show-ratio
                 (/ (- current-time console-last-state-change-time)
                    console-showing-time))))
      ((hiding)
       (if (>= current-time (+ console-last-state-change-time
                               console-hiding-time))
           (begin
             (set! console-last-state-change-time current-time)
             (set! console-state 'hidden)
             (set! console-show-ratio 0))
           (set! console-show-ratio
                 (- 1
                    (/ (- current-time console-last-state-change-time)
                       console-hiding-time)))))
      (else
       (error "console-update-state!: Unknown console state " console-state)))))

(define (console-start-showing!)
  (let ((current-time (milliseconds-since-startup)))
    (case console-state
      ((shown)
       ;; Make the current state last longer:
       (set! console-last-state-change-time current-time))
      ((showing)
       ;; Do nothing, the console is already showing:
       #t)
      ((hiding hidden)
       ;; Don't change the show ration, but change the scroll direction; we have to
       ;; "lie" about console-last-state-change-time:
       (set! console-last-state-change-time
             (- current-time
                (* console-showing-time console-show-ratio)))
       (set! console-state 'showing))
      (else
       (error "console-stard-showing!: Unknown console state " console-state)))))

(define (console-draw)
  ;; Compute the current console state, so that it's in the correct position when
  ;; we draw it:
  (console-update-state!)
  (unless (zero? console-show-ratio) ;; do nothing if the console is completely hidden
    (let* ((console-x1 console-external-border)
           (console-width (- (window-width) (* 2 console-external-border)))
           (console-height (+ (* console-line-no (font-height))
                              (* console-line-no console-inter-line-spacing)
                              (* 2 console-internal-border)))
           (console-y1 (- (window-height)
                          (* console-show-ratio
                             (+ console-external-border
                                console-height))))
           (first-line-x1 (+ console-x1 console-internal-border))
           (first-line-y1 (+ console-y1 console-internal-border)))
      (if console-background-color
          (draw-rectangle console-x1 console-y1 console-width console-height console-background-color))
      (console-draw-text console-lines first-line-x1 first-line-y1))))
  
(define (console-draw-text console-lines first-line-x1 first-line-y1)
  (if (null? console-lines)
      'done
      (begin
        (draw-text (car console-lines) first-line-x1 first-line-y1 console-foreground-color)
        (console-draw-text (cdr console-lines) first-line-x1 (+ first-line-y1
                                                                console-inter-line-spacing
                                                                (font-height))))))

(define (console-write-line! new-line)
  (console-start-showing!)
  (set! console-lines
        (append (if (= (length console-lines) console-line-no)
                    (cdr console-lines)
                    console-lines)
                (list new-line))))

(define (console-clear!)
  (set! console-lines '()))


(define (quit-event? event)
  (or (equal? event '(quit))
      ;;(equal? event '(key-down #\q))
      (equal? event '(key-down escape))))

(define (quit-event-handler event)
  (if (quit-event? event)
      (exit)
      (format #t "quit-event-handler: received the event ~s\n" event)))

;;; This is useful for simple tests, not for interactive-game; notice that
;;; this procedure handles all pending events (not just one) and it has
;;; zero parameters; hence it's *not* suitable to be used as a parameter
;;; of handle-all-enqueued-events:
(define (handle-quit-event-only)
  (handle-all-enqueued-events quit-event-handler))


;;; Handle all enqueued events with the given handler. Return #t if there
;;; was some event #f if there was none.
(define (handle-all-enqueued-events handle-one-event)
  (handle-all-enqueued-events-with-accumulator handle-one-event #f))
(define (handle-all-enqueued-events-with-accumulator handle-one-event did-we-see-some-event)
  (let ((next-event-or-#f (get-event)))
    (if next-event-or-#f
        (begin
          (handle-one-event next-event-or-#f)
          (handle-all-enqueued-events-with-accumulator handle-one-event #t))
        did-we-see-some-event)))

;;; This is useful as the last function call of test programs which are not loops:
(define (wait-till-the-user-quits)
  ;; Redraw the graphic window:
  (refresh)
  ;; Exit the program if the user has asked for it:
  (handle-quit-event-only)
  ;; Sleep for a short time, so that we don't use 100% of the processor time:
  (usleep 50)
  ;; Go again, in an infinite loop:
  (wait-till-the-user-quits))

;;; Support for different levels of detail. The best-looking levels are
;;; computation-intensive, so this is useful for slow computers:
(define draw-possibly-nuanced-circle draw-nuanced-circle)
(define detail-levels
  '(nuanced circle circle-border square square-border))
(define detail-level
  (car detail-levels))
(define (rotate-detail-level!)
  (set! detail-levels (append (cdr detail-levels)
                              (list (car detail-levels))))
  (set! detail-level (car detail-levels))
  (case detail-level
    ((nuanced)
     (set! draw-possibly-nuanced-circle draw-nuanced-circle))
    ((circle)
     (set! draw-possibly-nuanced-circle draw-circle))
    ((circle-border)
     (set! draw-possibly-nuanced-circle draw-circle-border))
    ((square)
     (set! draw-possibly-nuanced-circle draw-centered-square))
    ((square-border)
     (set! draw-possibly-nuanced-circle draw-centered-square-border))
    (else
     (error "Unknown detail level" detail-level))))

;;; Draw pills (each pill is simply represented by the cons of its two coordinates):
(define pill-color (make-color .5 1 .5))
(define (draw-pill pill)
  (draw-possibly-nuanced-circle (pill-x pill) (pill-y pill) pill-radius pill-color))

;; ;;; Draw the player:
;; (define left-eye-delta-angle 0.5)
;; (define right-eye-delta-angle (- left-eye-delta-angle))
;; (define eye-color (light (light yellow)))
;; (define avatar-name-color (make-color 1 1 1))
;; (define (draw-player x y radius angle name color)
;;   (let* ((left-eye-angle (+ angle left-eye-delta-angle))
;;          (right-eye-angle (+ angle right-eye-delta-angle))
;;          (tail-angle (+ angle pi))
;;          (sin-tail-angle (sin tail-angle))
;;          (cos-tail-angle (cos tail-angle))
;;          (tail-segment-1-distance (* radius 1.0))
;;          (tail-segment-2-distance (* radius 1.3))
;;          (tail-segment-3-distance (* radius 1.5))
;;          (tail-segment-4-distance (* radius 1.7))
;;          (eye-radius (/ radius 3))
;;          (left-eye-center-x (+ x (* radius (cos left-eye-angle))))
;;          (left-eye-center-y (- y (* radius (sin left-eye-angle))))
;;          (right-eye-center-x (+ x (* radius (cos right-eye-angle))))
;;          (right-eye-center-y (- y (* radius (sin right-eye-angle))))
;;          (tail-segment-1-center-x (+ x (* tail-segment-1-distance cos-tail-angle)))
;;          (tail-segment-1-center-y (- y (* tail-segment-1-distance sin-tail-angle)))
;;          (tail-segment-2-center-x (+ x (* tail-segment-2-distance cos-tail-angle)))
;;          (tail-segment-2-center-y (- y (* tail-segment-2-distance sin-tail-angle)))
;;          (tail-segment-3-center-x (+ x (* tail-segment-3-distance cos-tail-angle)))
;;          (tail-segment-3-center-y (- y (* tail-segment-3-distance sin-tail-angle)))
;;          (tail-segment-4-center-x (+ x (* tail-segment-4-distance cos-tail-angle)))
;;          (tail-segment-4-center-y (- y (* tail-segment-4-distance sin-tail-angle))))
;;     (draw-possibly-nuanced-circle tail-segment-4-center-x tail-segment-4-center-y (/ radius 6) color)
;;     (draw-possibly-nuanced-circle tail-segment-3-center-x tail-segment-3-center-y (/ radius 5) color)
;;     (draw-possibly-nuanced-circle tail-segment-2-center-x tail-segment-2-center-y (/ radius 4) color)
;;     (draw-possibly-nuanced-circle tail-segment-1-center-x tail-segment-1-center-y (/ radius 3) color)
;;     (draw-possibly-nuanced-circle left-eye-center-x left-eye-center-y eye-radius eye-color)
;;     (draw-possibly-nuanced-circle right-eye-center-x right-eye-center-y eye-radius eye-color)
;;     (draw-possibly-nuanced-circle x y radius color)
;;     (if name
;;         (draw-centered-text name x (+ y radius (font-height)) avatar-name-color))))

;;; Draw the player:
(define left-eye-delta-angle 0.5)
(define right-eye-delta-angle (- left-eye-delta-angle))
(define eye-color (light (light yellow)))
(define avatar-name-color (make-color 1 1 1))
(define tail-segment-no 3)
(define (draw-player x y radius angle name color)
  (let* ((left-eye-angle (+ angle left-eye-delta-angle))
         (right-eye-angle (+ angle right-eye-delta-angle))
         (tail-angle (+ angle pi))
         (sin-tail-angle (sin tail-angle))
         (cos-tail-angle (cos tail-angle))
         (eye-radius (/ radius 3))
         (left-eye-center-x (+ x (* radius (cos left-eye-angle))))
         (left-eye-center-y (- y (* radius (sin left-eye-angle))))
         (right-eye-center-x (+ x (* radius (cos right-eye-angle))))
         (right-eye-center-y (- y (* radius (sin right-eye-angle)))))
    (down-for i tail-segment-no 1
      (let ((segment-radius (* radius (/ 2 (+ i 2))))
            (segment-center-distance (* radius (+ i 1) 0.4)))
        (draw-possibly-nuanced-circle (+ x (* segment-center-distance cos-tail-angle))
                                      (- y (* segment-center-distance sin-tail-angle))
                                      segment-radius
                                      color)))
    (draw-possibly-nuanced-circle left-eye-center-x left-eye-center-y eye-radius eye-color)
    (draw-possibly-nuanced-circle right-eye-center-x right-eye-center-y eye-radius eye-color)
    (draw-possibly-nuanced-circle x y radius color)
    (if name
        (draw-centered-text name x (+ y radius (font-height)) avatar-name-color))))

;;; This file provides an implementation of the game state, used in the
;;; server.


;;; The module we include here is an implementation of records for Guile.
;;; It's explained very well in the manual:
(use-modules (srfi srfi-9))

;;; We need some constants defined in this module:

(use-modules (srfi srfi-9)) ;; this is needed for records

;;; The game field size, in pixels:
(define field-width  800)
(define field-height 600)

;;; The "border" size, i.e. the number of pixels within the field which are
;;; impossible to reach for players:
(define field-border 20)
(define minimum-x-in-the-field field-border)
(define minimum-y-in-the-field field-border)
(define maximum-x-in-the-field (- field-width field-border))
(define maximum-y-in-the-field (- field-height field-border))
(define (in-field-x? x)
  (and (<= minimum-x-in-the-field x)
       (<= x maximum-x-in-the-field)))
(define (in-field-y? y)
  (and (<= minimum-y-in-the-field y)
       (<= y maximum-y-in-the-field)))

;;; Player:
(define player-radius 12) ;12)
(define 2player-radius (* player-radius 2))

;;; How many updates per second to compute and to send to clients (by default).
;;; Notice that about ten decisions per seconds are more enough to play well;
;;; it's still faster than the human reaction time, and with a small frequency
;;; we don't stress the processor too much.
(define default-server-update-recomputation-frequency 62)
(define default-server-update-sending-frequency 30)
(define server-update-sending-frequency-for-client 10);60)
(define server-update-sending-frequency-for-ai-client 15)

;;; We represent the current state of each pill as a record:
(define-record-type pill
  (make-pill-by-specifying-all-fields x y velocity-x velocity-y)
  pill?
  (x                 pill-x                 pill-set-x!)
  (y                 pill-y                 pill-set-y!)
  (velocity-x        pill-velocity-x        pill-set-velocity-x!)
  (velocity-y        pill-velocity-y        pill-set-velocity-y!))

(define (make-pill)
  (make-pill-by-specifying-all-fields (random-x-in-the-field)
                                      (random-y-in-the-field)
                                      (random-from-to (- maximum-pill-component-velocity-in-modulo)
                                                      maximum-pill-component-velocity-in-modulo)
                                      (random-from-to (- maximum-pill-component-velocity-in-modulo)
                                                      maximum-pill-component-velocity-in-modulo)))

;;; Pills:
(define pills-no 20);30)
(define pill-radius 5)
(define pill-eating-radius (max (* player-radius 3/2)
                                (* pill-radius 3/2)))

;;; Maximum translation velocity in modulo for pills, for each component (x and y),
;;; in pixels per second:
(define maximum-pill-component-velocity-in-modulo
  50)

;;; Rotation velocity, in radiants per second:
(define rotation-velocity 9)

;;; Translation velocity, in pixels per second:
(define translation-velocity 200);300)

;;; Compute random reachable coordinates within the field:
(define (random-x-in-the-field)
  (random-from-to field-border maximum-x-in-the-field))
(define (random-y-in-the-field)
  (random-from-to field-border maximum-y-in-the-field))

;;; Saturate coordinates (meaning: return them unchanged if they are within
;;; the game field, otherwise return the appropriate border value):
(define (saturate-x x)
  (saturate x field-border maximum-x-in-the-field))
(define (saturate-y y)
  (saturate y field-border maximum-y-in-the-field))

;;; Compute the Euclidean distance from the given position to the given pill:
(define (distance-from-pill x y pill)
  (distance (pill-x pill) (pill-y pill) x y))

;;; Is the given pill within the eating range from the given position?
(define (eaten? x y pill)
  ;; Notice that a pill is represented as the cons of its coordinates:
  (<= (distance-from-pill x y pill)
      pill-eating-radius))

;;; Return the angle which must be added to the given angle of an object in the
;;; given position so that the object in the given position is oriented towards
;;; the given point:
(define (angle-to-target x y angle target-x target-y)
  (let* ((delta-x (- target-x x))
         (delta-y (- y target-y))
         (delta-x (if (zero? delta-x) 0.0001 delta-x)))
    (normalize-angle (- (if (> delta-x 0)
                            (atan (/ delta-y delta-x))
                            (+ pi (atan (/ delta-y delta-x))))
                        angle))))

;;; Note that there are several global variables in this source file
;;; which are used in an imperative way, and we need to protect them
;;; from concurrent access with a mutex. Such synchronization is *not*
;;; implemented here, and it's the caller's responsibility.

;;; We represent the current state of each player (we can have arbitrarily
;;; many players in this particular game) as a record:
(define-record-type player
  (make-player-by-specifying-all-fields
      name
      red green blue
      score
      x
      y
      angle ; in radiants
      translation-state ; one of: forward, backward, idle
      rotation-state ; one of: left, right, idle
      update-frequency ; how many updates per second the client wants
      changed-update-frequency) ; has the update frequency just changed?
  player?
  (name              player-name              player-set-name!)
  (red               player-red               player-set-red!)
  (green             player-green             player-set-green!)
  (blue              player-blue              player-set-blue!)
  (score             player-score             player-set-score!)
  (x                 player-x                 player-set-x!)
  (y                 player-y                 player-set-y!)
  (angle             player-angle             player-set-angle!)
  (translation-state player-translation-state player-set-translation-state!)
  (rotation-state    player-rotation-state    player-set-rotation-state!)
  (update-frequency  player-update-frequency  player-set-update-frequency!)
  (changed-update-frequency  player-changed-update-frequency player-set-changed-update-frequency!))

;;; A procedure for making a player without specifying a value for all the 
;;; fields:
(define (make-player name)
  (make-player-by-specifying-all-fields name
                               (random 1.0) (random 1.0) (random 1.0)
                               0
                               (random-x-in-the-field)
                               (random-y-in-the-field)
                               (normalize-angle (random 2pi))
                               'idle
                               'idle
                               default-server-update-sending-frequency
                               #f))

;;; Add a new player to the game:
(define (add-player! player)
  (set! players (cons player players)))

;;; Remove a player from the game:
(define (remove-player! player)
  (set! players (without-by-identity player players)))

;;; Add the given number of new pills:
(define (add-pills! how-many)
  (set! pills (append (map (lambda whatever
                             (make-pill))
                           (iota how-many))
                      pills)))

;;; Remove an eaten pill:
(define (remove-pill! pill)
  (set! pills (without-by-identity pill pills)))

;;; Compute the next game state, imperatively set the global variables and
;;; return a message with state changes:
(define (update-state! seconds-since-the-last-update)
  (update-pill-state! seconds-since-the-last-update)
  (iterate (lambda (player)
             (update-player-state! player seconds-since-the-last-update))
           players)
  ;; Return the update message:
  (list  'updated-state
         (map (lambda (p)
                (list (player-name p) (player-score p)
                      (player-x p) (player-y p) (player-angle p)
                      (player-red p) (player-green p) (player-blue p)))
              players)
         (map (lambda (pill)
                (cons (pill-x pill) (pill-y pill)))
              pills)))

;;; Update the state of all the currently-existing pills:
(define (update-pill-state! delta-t) ; delta-t is in seconds
  (iterate (lambda (pill)
             (let* ((new-x (+ (pill-x pill)
                              (* (pill-velocity-x pill) delta-t)))
                    (new-y (+ (pill-y pill)
                              (* (pill-velocity-y pill) delta-t)))
                    (bounced-x (not (in-field-x? new-x)))
                    (bounced-y (not (in-field-y? new-y))))
               (pill-set-x! pill (saturate-x new-x))
               (pill-set-y! pill (saturate-y new-y))
               (when bounced-x
                 (pill-set-velocity-x! pill (- (pill-velocity-x pill))))
               (when bounced-y
                 (pill-set-velocity-y! pill (- (pill-velocity-y pill))))))
           pills))

;;; Update the state of a player:
(define (update-player-state! player delta-t) ; delta-t is in seconds
  ;; Update angle, x and y according to the current states:
  (let* ((old-angle (player-angle player))
         (old-x (player-x player))
         (old-y (player-y player))
         (rotation-quotient (case (player-rotation-state player)
                              ((left) 1)
                              ((right) -1)
                              ((idle) 0)))
         (translation-quotient (case (player-translation-state player)
                                 ((forward) 1)
                                 ((backward) -1)
                                 ((idle) 0)))
         (new-angle (normalize-angle (+ old-angle
                                        (* rotation-quotient
                                           rotation-velocity
                                           delta-t))))
         (new-x (saturate-x (+ old-x (* translation-quotient
                                        (cos new-angle)
                                        translation-velocity
                                        delta-t))))
         (new-y (saturate-y (- old-y (* translation-quotient
                                        (sin new-angle)
                                        translation-velocity
                                        delta-t))))
         (eaten-pills (filter (lambda (pill) (eaten? new-x new-y pill))
                              pills))
         (eaten-pills-no (length eaten-pills))
         (old-score (player-score player))
         (new-score (+ old-score eaten-pills-no)))
    (player-set-angle! player new-angle)
    ;; Don't move if the player would collide with any other player:
    (unless (exists? (lambda (other-player)
                       (< (distance new-x new-y
                                    (player-x other-player) (player-y other-player))
                          2player-radius))
                     (without-by-identity player players))
      (player-set-x! player new-x)
      (player-set-y! player new-y))
    (player-set-score! player new-score)
    (set! pills (subtract-lists-by-identity pills eaten-pills))
    (add-pills! eaten-pills-no)))

;;; The current list of players, as an unsorted list:
(define players
  '())

;;; The current list of pills, as an unsorted list:
(define pills
  (map (lambda whatever (make-pill))
       (iota pills-no)))

(randomize-random-state!)

(define p
  (let ((user-name (getenv "USER")))
    (make-player user-name)))
(add-player! p)

(define bots-no 0)
(define bots
  (map (lambda (index)
         (make-player (string-append "bot" (number->string (+ index 1))) #f))
       (iota bots-no)))
(iterate add-player! bots)

(define (handle-event event)
  (cond ((quit-event? event)
         (exit 0))
        ((equal? event '(key-down #\b))
         (let ((new-bot (make-player (string-append "bot" (number->string (+ (length bots) 1))))))
           (set! bots (cons new-bot bots))
           (add-player! new-bot)))
        ((equal? event '(key-down left))
         (player-set-rotation-state! p 'left))
        ((equal? event '(key-down right))
         (player-set-rotation-state! p 'right))
        ((or (equal? event '(key-up left))
             (equal? event '(key-up right)))
         (player-set-rotation-state! p 'idle))
        ((equal? event '(key-down up))
         (player-set-translation-state! p 'forward))
        ((equal? event '(key-down down))
         (player-set-translation-state! p 'backward))
        ((equal? event '(key-down #\d))
         (rotate-detail-level!))
        ((or (equal? event '(key-up up))
             (equal? event '(key-up down)))
         (player-set-translation-state! p 'idle))))

(define pills
  (map (lambda (whatever)
         (make-pill))
       (iota pills-no)))

(define (nearest-pill x y)
  (define (nearest-pill pills best-distance-till-now best-pill-till-now)
    (if (null? pills)
        best-pill-till-now
        (let* ((first-pill (car pills))
               (first-pill-distance (distance-from-pill x y first-pill)))
          (if (< first-pill-distance best-distance-till-now)
              (nearest-pill (cdr pills) first-pill-distance first-pill)
              (nearest-pill (cdr pills) best-distance-till-now best-pill-till-now)))))
  (nearest-pill (cdr pills) (distance-from-pill x y (car pills)) (car pills)))

(define (update-bot-state! bot)
  (let* ((x (player-x bot))
         (y (player-y bot))
         (angle (player-angle bot))
         (nearest-pill (nearest-pill x y))
         (angle-to-target (angle-to-target x y angle (pill-x nearest-pill) (pill-y nearest-pill))))
    (if (> (abs angle-to-target) 0.3)
        (begin
          (player-set-rotation-state! bot (if (> angle-to-target 0) 'left 'right))
          (player-set-translation-state! bot 'idle))
        (begin
          (player-set-rotation-state! bot 'idle)
          (player-set-translation-state! bot 'forward)))))

(define (work delta-t)
  (handle-all-enqueued-events handle-event)
  (update-state! delta-t)
  (iterate update-bot-state! bots)
  (clear black)
  (iterate (lambda (p)
             (draw-player (player-x p)
                          (player-y p)
                          player-radius
                          (player-angle p)
                          (string-append (player-name p) " (" (number->string (player-score p)) ")")
                          (make-color (player-red p)
                                      (player-green p)
                                      (player-blue p))))
           players)
  (iterate draw-pill pills)
  (console-update-state!)
  (console-draw)
  (refresh))

(call-with-maximum-frequency 70 work)
