;;; This file is part of guile-game, a graphic library for Guile

;;; Copyright (C) 2009, 2010  Luca Saiu

;;; This program is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;; The files in this directory are the actual Scheme sources of my
;; implementation, which my students couldn't originally access; their
;; task consisted in implementing similar funcionality, so I made it
;; (well, let's say relatively) hard to look at what I did.
;;   Everything is realeased under the GPLv3-or-later at this point,
;; if you are reading this.


;;; This file contains entire coies of files in the scm/ directory: relative
;;; paths with load don't play well with source scrambling.

(initialize-graphics "Pong monitor" 600 600)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;; From color.scm:

;;; I pre-define some useful colors. Of course, always remember that color
;;; coordinates are numbers (floats like 0.7, but also also *rationals* 
;;; like 1/3 are accepted in Scheme) between 0 and 1:
(define black   (make-color 0   0   0))
(define white   (make-color 1/2 1/2 1/2))
(define red     (make-color 1/2 0   0))
(define green   (make-color 0   1/2 0))
(define blue    (make-color 0   0   1/2))
(define cyan    (make-color 0   1/2 1/2))
(define magenta (make-color 1/2 0   1/2))
(define yellow  (make-color 1/2 1/2 0))
(define gray    (make-color 1/2 1/2 1/2))

;;; Return a new color obtained by multiplying each component of the given
;;; color by the given factor, and saturating:
(define (color-* factor color)
  (let ((red   (color->red color))
        (green (color->green color))
        (blue  (color->blue color)))
    (make-color (saturate (* factor red) 0 1)
                (saturate (* factor green) 0 1)
                (saturate (* factor blue) 0 1))))

;;; Given a color, return a ligher version of it. For example (light green)
;;; makes a lighter green:
(define (light color)
  (color-* 3/2 color))

;;; Given a color, return a darker version of it. For example (dark green)
;;; makes a darker green:
(define (dark color)
  (color-* 2/3 color))

;; Return a new random color:
(define (random-color)
  (make-color (random 1.0) (random 1.0) (random 1.0)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;; From utility.scm:

;; This is the most general version of for, but it's not always
;; comfortable to call. I use this to define three friendlier
;; versions:
(define-macro (for-step variable from to step . body)
  (let ((to-name (gensym))
        (from-name (gensym))
        (step-name (gensym))
        (lower-limit-name (gensym))
        (upper-limit-name (gensym)))
    `(let* ((,from-name ,from)
            (,to-name ,to)
            (,variable ,from-name)
            (,lower-limit-name (min ,from-name ,to-name))
            (,upper-limit-name (max ,from-name ,to-name))
            (,step-name ,step))
       (while (and (>= ,variable ,lower-limit-name)
                   (<= ,variable ,upper-limit-name))
              ,@(append body
                        `((set! ,variable (+ ,variable ,step-name))))))))

;; Evaluate the body a number of times equal to how-many-times.
;; No iteration variable is visible in the body,
;;
;; Example: (n-times 5 (display "This will be printed five times\n")
(define-macro (n-times how-many-times . body)
  (let ((variable (gensym)))
    `(for ,variable 1 ,how-many-times ,@body)))

;; The classic for loop, with the variable incremented at each iteration.
;;
;; for example (for i 1 5
;;               (display "i is now ")
;;               (display i)
;;               (newline))
;; is equivalent to the C loop:
;; { 
;;   int i;
;;   for(i = 1; i <= 5; i++){
;;     printf("i is now ");
;;     printf("%i", i);
;;     printf("\n");
;;   } // for
;; }
(define-macro (for variable from to . body)
  `(for-step ,variable ,from ,to 1 ,@body))

;; Just like for, but here the variable is *de*cremented at each iteration:
(define-macro (down-for variable from to . body)
  `(for-step ,variable ,from ,to -1 ,@body))

;;; Return the n-th element of the given list:
(define (n-th n list)
  (if (zero? n)
      (car list)
      (n-th (- n 1)
            (cdr list))))

;;; Convert the given s-expression into its printer representation:
(define (s-expression->string s-expression)
  (let ((string-port (open-output-string)))
    (write s-expression string-port)
    (get-output-string string-port)))

;;; Some useful constants:
(define pi (* 2 (acos 0)))
(define -pi (- pi))
(define 2pi (* 2 pi))
(define pi/2 (/ pi 2))
(define pi/3 (/ pi 3))
(define pi/4 (/ pi 4))
(define pi/6 (/ pi 6))

;;; Given an angle in radiants, return another equivalent one in (-pi, +pi]
(define (normalize-angle angle)
  (cond ((> angle pi)
         (normalize-angle (- angle 2pi)))
        ((<= angle -pi)
         (normalize-angle (+ angle 2pi)))
        (else
         angle)))

;;; Destructively update a variable representing an angle in radiants,
;;; replacing them with the normalized value:
(define-macro (normalize-angle! variable)
  `(set! ,variable (normalize-angle ,variable)))

;;; Convert an angle from radiants to degrees. This is probably only useful for debugging:
(define (radiants->degrees angle)
  (* 180 (/ angle pi)))

;;; Return a copy of the given list without the given element (if it is in the
;;; original list):
(define (without-with-equality-predicate equality-predicate element list)
  (define (without element list accumulator)
    (cond ((null? list)
           accumulator)
          ((equality-predicate element (car list))
           (append (cdr list) accumulator))
          (else
           (without element (cdr list) (cons (car list) accumulator)))))
  (without element list '()))

;;; The "default" function simply called without uses equal? (structural comparison),
;;; and is just an alias for without-by-structure:
(define (without-by-structure element list)
  (without-with-equality-predicate equal? element list))
(define without without-by-structure)
 
;;; without-by-identity uses eq? (equality by identity)
(define (without-by-identity element list)
  (without-with-equality-predicate eq? element list))

;;; Return a copy of the first parameter without the elements in the second parameter.
;;; This is an implementation of set subtraction, where; sets are implemented as unsorted lists.
(define (subtract-lists-with-equality-function equality-function? list elements-to-remove)
  (if (null? elements-to-remove)
      list
      (subtract-lists-with-equality-function equality-function?
                                             (without-by-structure (car elements-to-remove) list)
                                             (cdr elements-to-remove))))
(define (subtract-lists-by-identity list elements-to-remove)
  (subtract-lists-with-equality-function eq? list elements-to-remove))
(define (subtract-lists-by-structure list elements-to-remove)
  (subtract-lists-with-equality-function equal? list elements-to-remove))

;;; Call the given function on the elements of the given list, in order. The results
;;; are ignored:
(define (iterate function list)
  (if (null? list)
      #t
      (begin
        (function (car list))
        (iterate function (cdr list)))))

;;; Common Lisp-style when and unless. These are one-way if conditionals, which take
;;; one or more forms as their body even without using begin. when executes its body
;;; (and returns the value of its last form) if the condition is true, and unless
;;; executes its body (and returns the value of its last form) if the condition is
;;; false:
(define-macro (when condition . body)
  `(if ,condition
       (begin
         ,@body)))
(define-macro (unless condition . body)
  `(if (not ,condition)
       (begin
         ,@body)))

;;; Useful higher-order functions working on lists (note that map and filter
;;; are predefined):
(define (for-all? predicate? list)
  (cond ((null? list)
         #t)
        ((predicate? (car list))
         (for-all? predicate? (cdr list)))
        (else
         #f)))
(define (exists? predicate? list)
  (cond ((null? list)
         #f)
        ((predicate? (car list))
         #t)
        (else
         (exists? predicate? (cdr list)))))

;;; Executes the forms in body and return the result of the last one. If any
;;; of them throws an exception, just capure it, stop executing the body and
;;; return #f:
(define-macro (ignore-failures . body)
  `(catch #t
     (lambda ()
       ,@body)
     (lambda exception
       (format #t "ignore-failures: ignored the exception ~s\n" exception)
       #f)))

(define (square x)
  (* x x))

;;; Compute the Euclidean distance between two points:
(define (distance x1 y1 x2 y2)
  (sqrt (+ (square (- x1 x2))
           (square (- y1 y2)))))

;;; Set the initial random state to a value depending on the current time,
;;; so that every time we start up the program the sequence of random values
;;; is different:
(define (randomize-random-state!)
  (let ((time (gettimeofday)))
    (set! *random-state*
          (seed->random-state (+ (car time)
                                 (cdr time)
                                 (getpid))))))

;;; This is very useful for debugging. The given expression is evaluated and printed
;;; on the output in a clearly readable format. For example (log (+ 1 2)) prints
;;; "(+ 1 2) ==> 3" on the output.
(define-macro (log expression)
  `(format #t "~s ==> ~s\n" ',expression ,expression))

;;; Call the given procedure in a loop, at most maximum-frequency times per second
;;; -- or less frequently if the machine is not fast enough. Exit the loop if the
;;; function returns #f.
;;; The called procedure is given the time elapsed since the last call in seconds
;;; as its only parameter.
(define (call-with-maximum-frequency-till-#f maximum-frequency ; maximum-frequency is in Hz
                                             procedure)
  (let* ((minimum-period (/ 1000 maximum-frequency)) ; in milliseconds
         (minimum-period-in-seconds (/ minimum-period 1000))
         (time-before-last-period (milliseconds-since-startup))
         (should-we-go-on #t))
    (while should-we-go-on
      (let* ((time-before-this-period (milliseconds-since-startup))
             (delta-t (- time-before-this-period
                         time-before-last-period))
             (milliseconds-to-sleep (- minimum-period delta-t)))
        (if (> milliseconds-to-sleep 0)
            (begin
              (usleep (inexact->exact (floor (* milliseconds-to-sleep 1000))))
              (set! time-before-last-period (+ time-before-this-period
                                               milliseconds-to-sleep))
              (set! should-we-go-on
                    (procedure minimum-period-in-seconds)))
            (begin
              ;;(usleep 0) ; this should at least yield to another process
              (set! time-before-last-period time-before-this-period)
              (set! should-we-go-on
                    (procedure (/ delta-t 1000)))))))))

;;; Call the given procedure in an infinite loop, at most maximum-frequency times
;;; per second -- or less frequently if the machine is not fast enough.
;;; The called procedure is given the time elapsed since the last call in seconds
;;; as its only parameter.
;;;
;;; If the given procedure returns a number, that is interpreted as the new
;;; frequency to be used from that point on.
(define (call-with-maximum-frequency maximum-frequency ; maximum-frequency is in Hz
                                     procedure)
  (let ((current-maximum-frequency maximum-frequency))
    (while #t
      (call-with-maximum-frequency-till-#f
          current-maximum-frequency
          (lambda (time-since-last-call)
            (let ((new-frequency-or-nonnumber (procedure time-since-last-call)))
              (if (number? new-frequency-or-nonnumber)
                (begin
                  (set! current-maximum-frequency new-frequency-or-nonnumber)
                  #f)     ; stop; a new loop will begin with the new frequency
                #t))))))) ; loop again

;; (define original-display
;;   (if (defined? 'original-display)
;;      original-display
;;      display))
;; (define (display . arguments)
;;   (original-display (get-internal-real-time))
;;   (original-display " ")
;;   (apply original-display arguments)
;;   (flush-all-ports)
;;   ;(usleep 1000)
;;   )

(define (fatal . arguments)
  (format #t "Fatal error: ~s\nExiting...\n" arguments)
  (kill-all-threads))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;; From graphics.scm:
;;; Draw a *non-filled* ractangle. The interface is compatible with
;;; draw-rectangle:
(define (draw-rectangle-border x0 y0 width height color)
  (draw-line x0 y0 (+ x0 width -1) y0 color)
  (draw-line x0 y0 x0 (+ y0 height -1) color)
  (draw-line x0 (+ y0 height -1) (+ x0 width -1) (+ y0 height -1) color)
  (draw-line (+ x0 width -1) y0 (+ x0 width -1) (+ y0 height -1) color))

(define (draw-block x0 y0 width height color . border-size-or-null)
  (let ((width/2 (/ width 2))
        (height/2 (/ height 2)))
    (apply draw-noncentered-block
           (append (list (- x0 width/2)
                         (- y0 height/2)
                         width
                         height
                         color)
                   border-size-or-null))))

;;; Draw a block with lights and shadows, suitable for Tetris and Breakout;
;;; notice that there is an optional parameter (which is seen by the procedure
;;; as an empty list when not given, and a non-empty list when given):
(define (draw-noncentered-block x0 y0 width height color . border-size-or-null)
  ;; If the optional parameter was supplied then bind border-size to the
  ;; value given by the user, otherwise use a default:
  (let ((border-size (if (null? border-size-or-null)
                         3
                         (car border-size-or-null))))
    ;; Draw nothing if the block is too small:
    (if (or (<= width (* 2 border-size))
            (<= height (* 2 border-size)))
        #t ; well, we have to return something, so we return #t
        ;; Ok, the block is not too small: draw it:
        (let ((lighter-color (light color)) ; color for the lightened part
              (darker-color (dark color)))  ; color for the shadowed part
          ;; Draw the shadowed part on the East and South sides:
          (draw-rectangle x0 (+ y0 height (- border-size)) width border-size darker-color)
          (draw-rectangle (+ x0 width (- border-size)) y0 border-size height darker-color)
          ;; Draw the enlightended part on the West and Nort sides:
          (draw-rectangle x0 y0 border-size height lighter-color)
          (draw-rectangle x0 y0 width border-size lighter-color)
          ;; Draw the inside part:
          (draw-rectangle (+ x0 border-size) (+ y0 border-size)
                          (- width (* 2 border-size)) (- height (* 2 border-size))
                          color))))) ;; There's no automatic refresh here!

(define (square a)
  (* a a))

;;; The absolute value should not be needed here, but I've found that some rounding
;;; errors with floats make the argument of the square root be negative, and the result
;;; of the square root complex...
;;; If you use this procedure to draw large circles the result will not look good around
;;; the leftmost and rightmost points. But it is reasonably fast, to be used for
;;; interactive applications where speed matters. The result could be made better-looking
;;; by adding another loop on the y variable.
(define (draw-circle center-x center-y radius color)
  (for x (- center-x radius) (+ center-x radius)
       (let ((delta-y (sqrt (abs (- (square radius)
                                    (square (- x center-x)))))))
         (draw-line x (- center-y delta-y) x (+ center-y delta-y) color))))
;;; See the comment above about abs. But notice that this procedure gives good results
;;; also for big circles.
(define (draw-circle-border center-x center-y radius color)
  (for x (- center-x radius) (+ center-x radius)
       (let ((delta-y (sqrt (abs (- (square radius)
                                    (square (- x center-x)))))))
         (draw-pixel x (- center-y delta-y) color)
         (draw-pixel x (+ center-y delta-y) color))))

(define nuance-step -2) ;; Make this bigger in modulo (and still negative) for slow machines
(define (draw-nuanced-circle center-x center-y maximum-radius color)
  (for-step radius maximum-radius 2 nuance-step
    (draw-circle center-x center-y radius
                 (color-* (- 1.2 (/ radius maximum-radius))
                          color))))

(define sqrt2 (sqrt 2))
;;; These can be used as an "approximation" of a circle on slow machines, for interactive
;;; applications which need speed. There is some ugly code duplication here, but the
;;; point of this code is speed, so I don't want to use a higher-order function:
(define (draw-centered-square-border center-x center-y half-diagonal color)
  (let* ((half-side (/ half-diagonal sqrt2))
         (side (* half-side 2))
         (top-left-x (- center-x half-side))
         (top-left-y (- center-y half-side)))
    (draw-rectangle-border top-left-x top-left-y side side color)))
(define (draw-centered-square center-x center-y half-diagonal color)
  (let* ((half-side (/ half-diagonal sqrt2))
         (side (* half-side 2))
         (top-left-x (- center-x half-side))
         (top-left-y (- center-y half-side)))
    (draw-rectangle top-left-x top-left-y side side color)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;; From console.scm:
;;; The user should *not* load this file directly. It's already loaded by
;;; interactive.scm.

(define console-external-border 10)
(define console-internal-border 5)
(define console-line-no 5)
(define console-inter-line-spacing 2)

;;; If the background color is #f instead of a color, then the console is drawn
;;; with no background. This variable can be safely changed with set! after
;;; initialization:
(define console-background-color #f)
;(define console-background-color (make-color 0.2 0.0 0.0))

(define console-foreground-color (make-color 1 1 1))
;;(define console-foreground-input-color (light yellow))

;;; The console state can be one of of hidden, shown, showing or hiding.
(define console-state
  'hidden)

;;; A ratio of 1 means that the console is fully shown, a ratio of 0 means that
;;; it is fully hidden. All values in [0, 1] are valid.
(define console-show-ratio 0)

;;; The absolute time (as returned by milliseconds-since-startup) when the
;;; state changed the last time:
(define console-last-state-change-time 0)

(define console-full-showing-time-after-each-change 3000) ; in milliseconds

;;; How long the console takes, in milliseconds, for showing up or hiding:
(define console-showing-time 250);;500)
(define console-hiding-time 2000)

;;; The list of lines currently shown in the console, from the oldest to the newest,
;;; as a list of strings
(define console-lines
  '())
;;'("abc" "123" "foo bar baaz quux" "boo" "this is the bottom line"))

(define (console-update-state!)
  (let ((current-time (milliseconds-since-startup)))
    (case console-state
      ((hidden)
       ;; Do nothing
       #t)
      ((shown)
       (if (>= current-time (+ console-last-state-change-time
                               console-full-showing-time-after-each-change))
           (begin
             (set! console-last-state-change-time current-time)
             (set! console-state 'hiding))))
      ((showing)
       (if (>= current-time (+ console-last-state-change-time
                               console-showing-time))
           (begin
             (set! console-last-state-change-time current-time)
             (set! console-state 'shown)
             (set! console-show-ratio 1))
           (set! console-show-ratio
                 (/ (- current-time console-last-state-change-time)
                    console-showing-time))))
      ((hiding)
       (if (>= current-time (+ console-last-state-change-time
                               console-hiding-time))
           (begin
             (set! console-last-state-change-time current-time)
             (set! console-state 'hidden)
             (set! console-show-ratio 0))
           (set! console-show-ratio
                 (- 1
                    (/ (- current-time console-last-state-change-time)
                       console-hiding-time)))))
      (else
       (error "console-update-state!: Unknown console state " console-state)))))

(define (console-start-showing!)
  (let ((current-time (milliseconds-since-startup)))
    (case console-state
      ((shown)
       ;; Make the current state last longer:
       (set! console-last-state-change-time current-time))
      ((showing)
       ;; Do nothing, the console is already showing:
       #t)
      ((hiding hidden)
       ;; Don't change the show ration, but change the scroll direction; we have to
       ;; "lie" about console-last-state-change-time:
       (set! console-last-state-change-time
             (- current-time
                (* console-showing-time console-show-ratio)))
       (set! console-state 'showing))
      (else
       (error "console-stard-showing!: Unknown console state " console-state)))))

(define (console-draw)
  ;; Compute the current console state, so that it's in the correct position when
  ;; we draw it:
  (console-update-state!)
  (unless (zero? console-show-ratio) ;; do nothing if the console is completely hidden
    (let* ((console-x1 console-external-border)
           (console-width (- (window-width) (* 2 console-external-border)))
           (console-height (+ (* console-line-no (font-height))
                              (* console-line-no console-inter-line-spacing)
                              (* 2 console-internal-border)))
           (console-y1 (- (window-height)
                          (* console-show-ratio
                             (+ console-external-border
                                console-height))))
           (first-line-x1 (+ console-x1 console-internal-border))
           (first-line-y1 (+ console-y1 console-internal-border)))
      (if console-background-color
          (draw-rectangle console-x1 console-y1 console-width console-height console-background-color))
      (console-draw-text console-lines first-line-x1 first-line-y1))))
  
(define (console-draw-text console-lines first-line-x1 first-line-y1)
  (if (null? console-lines)
      'done
      (begin
        (draw-text (car console-lines) first-line-x1 first-line-y1 console-foreground-color)
        (console-draw-text (cdr console-lines) first-line-x1 (+ first-line-y1
                                                                console-inter-line-spacing
                                                                (font-height))))))

(define (console-write-line! new-line)
  (console-start-showing!)
  (set! console-lines
        (append (if (= (length console-lines) console-line-no)
                    (cdr console-lines)
                    console-lines)
                (list new-line))))

(define (console-clear!)
  (set! console-lines '()))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;; From interactive.scm:
(define (quit-event? event)
  (or (equal? event '(quit))
      ;;(equal? event '(key-down #\q))
      (equal? event '(key-down escape))))

(define (quit-event-handler event)
  (if (quit-event? event)
      (exit)
      (format #t "quit-event-handler: received the event ~s\n" event)))

;;; This is useful for simple tests, not for interactive-game; notice that
;;; this procedure handles all pending events (not just one) and it has
;;; zero parameters; hence it's *not* suitable to be used as a parameter
;;; of handle-all-enqueued-events:
(define (handle-quit-event-only)
  (handle-all-enqueued-events quit-event-handler))


;;; Handle all enqueued events with the given handler. Return #t if there
;;; was some event #f if there was none.
(define (handle-all-enqueued-events handle-one-event)
  (handle-all-enqueued-events-with-accumulator handle-one-event #f))
(define (handle-all-enqueued-events-with-accumulator handle-one-event did-we-see-some-event)
  (let ((next-event-or-#f (get-event)))
    (if next-event-or-#f
        (begin
          (handle-one-event next-event-or-#f)
          (handle-all-enqueued-events-with-accumulator handle-one-event #t))
        did-we-see-some-event)))

;;; This is useful as the last function call of test programs which are not loops:
(define (wait-till-the-user-quits)
  ;; Redraw the graphic window:
  (refresh)
  ;; Exit the program if the user has asked for it:
  (handle-quit-event-only)
  ;; Sleep for a short time, so that we don't use 100% of the processor time:
  (usleep 50)
  ;; Go again, in an infinite loop:
  (wait-till-the-user-quits))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Some other recycled code:
(define (unbind-function k a)
  (cond ((null? a) a)
        ((eq? (caar a) k) (cdr a))
        (else (cons (car a)
                    (unbind-function k (cdr a))))))
(define-macro (unbind k a) `(unbind-function ',k ,a))
(define (bind-function k v a)
  (acons k v (unbind-function k a)))
(define-macro (bind1 k v a) `(acons ',k ,v (unbind ,k ,a)))
(define-macro (bind . bindings-and-a)
  (if (null? (cdr bindings-and-a))
      (car bindings-and-a)
      `(bind1 ,(car bindings-and-a)
              ,(cadr bindings-and-a)
              (bind ,@(cddr bindings-and-a)))))
(define (lookup-function k a)
  (cond ((null? a)
         (error "lookup-function: unbound key" k))
        ((eq? (caar a) k)
         (cdar a))
        (else
         (lookup-function k (cdr a)))))
(define-macro (lookup k a)
  `(lookup-function ',k ,a))
(define (bound-function? k a)
  (cond ((null? a)
         #f)
        ((eq? (caar a) k)
         #t)
        (else
         (bound-function? k (cdr a)))))
(define-macro (bound? k a)
  `(bound-function? ',k ,a))

;;; Return x if it's within the given bounds (included), otherwise return the
;;; limit:
(define (saturate x minimum maximum)
  (cond ((< x minimum)
         minimum)
        ((> x maximum)
         maximum)
        (else
         x)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Here begins the actual monitor code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(define (random-x)
  (random-from-to 20 500))
(define (random-y)
  (random-from-to 20 500))

(define ball-color (light yellow))
(define paddle-color red)
;; (while #t
;;   (handle-quit-event-only)
;;   (clear black)
;;   (draw-block (random-x) (random-y) block-width block-height block-color)
;;   (refresh))

(set! console-background-color (dark blue))
(randomize-random-state!)

(define (draw-state state)
  (let ((balls (lookup balls state))
        (players (map cdr (lookup players state))))
    ;; (if (zero? (random 5000))
    ;;     (console-write-line! (string-append "round " (number->string (lookup round state)))))
    (clear black)
    (iterate (lambda (player)
               (draw-block (lookup x player)
                           (lookup y player)
                           (lookup width player)
                           (lookup height player)
                           paddle-color))
             players)
    (iterate (lambda (ball)
               (draw-nuanced-circle (lookup x ball)
                                    (lookup y ball)
                                    (/ (lookup width ball) 2)
                                    ball-color))
             balls)
    (iterate console-write-line! (lookup messages state))
    (console-draw)
    (refresh)))

(define (decide-command state)
  '(i-do-not-need-anything-thanks))

(define (handle-event-in-state event state)
  (cond ((quit-event? event)
         (exit))
        (else
         #t)))

(while #t
  (let ((state (read)))
    (if (eof-object? state)
        (exit -1)
        (begin
          ;; Decide and answer, immediately, so that we don't slow
          ;; down the server too much:          
          (let ((command (decide-command state)))
            (write command)
            (newline)
            (force-output))
          ;; Now we can think about graphics; and it's not a problem if we draw
          ;; the *previous* state:
          (handle-all-enqueued-events (lambda (event)
                                        (handle-event-in-state event state)))
          (draw-state state)))))
