;;; This file is part of guile-game, a graphic library for Guile

;;; Copyright (C) 2009, 2010  Luca Saiu

;;; This program is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;; The files in this directory are the actual Scheme sources of my
;; implementation, which my students couldn't originally access; their
;; task consisted in implementing similar funcionality, so I made it
;; (well, let's say relatively) hard to look at what I did.
;;   Everything is realeased under the GPLv3-or-later at this point,
;; if you are reading this.


(use-modules (ice-9 popen))

(define left 20)
(define top 20)
(define window-width 600)
(define window-height 600)
(define right (- window-width left))
(define bottom (- window-height (* 2 top) 120))

(define ball-radius 12)
(define ball-width (* ball-radius 2))
(define ball-height ball-width)
(define paddle-width 70)
(define paddle-height 20)
(define player-dx-modulo 2.5)
(define ball-velocity-modulo 3)
(define ball-no 3)

;;; Call the given function on the elements of the given list, in order. The results
;;; are ignored:
(define (iterate function list)
  (if (null? list)
      #t
      (begin
        (function (car list))
        (iterate function (cdr list)))))

;;; Return a copy of the given list without the given element (if it is in the
;;; original list):
(define (without-with-equality-predicate equality-predicate element list)
  (define (without element list accumulator)
    (cond ((null? list)
           accumulator)
          ((equality-predicate element (car list))
           (append (cdr list) accumulator))
          (else
           (without element (cdr list) (cons (car list) accumulator)))))
  (without element list '()))

;;; The "default" function simply called without uses equal? (structural comparison),
;;; and is just an alias for without-by-structure:
(define (without-by-structure element list)
  (without-with-equality-predicate equal? element list))
(define without without-by-structure)
 
;;; without-by-identity uses eq? (equality by identity)
(define (without-by-identity element list)
  (without-with-equality-predicate eq? element list))

(define (random-integer-between a b)
  "Return a random integer in [a, b]"
  (+ a (random (- b a -1))))
(define (random-float-between a b)
  "Return a random float in [a, b)"
  (let ((width (- b a))
        (ratio (random 1.0)))
    (+ a (* ratio width))))
(define (random-between minimum maximum)
  (if (and (integer? minimum)
           (integer? maximum))
      (random-integer-between minimum maximum)
      (random-float-between minimum maximum)))
(define (random-from-to minimum maximum)
  (random-between minimum maximum))

;;; Return x if it's within the given bounds (included), otherwise return the
;;; limit:
(define (saturate x minimum maximum)
  (cond ((< x minimum)
         minimum)
        ((> x maximum)
         maximum)
        (else
         x)))

(define (within? x minimum maximum)
  (and (>= x minimum)
       (<= x maximum)))

;;;; My alternate alist functions:
(define (unbind-function k a)
  (cond ((null? a) a)
        ((eq? (caar a) k) (cdr a))
        (else (cons (car a)
                    (unbind-function k (cdr a))))))
(define-macro (unbind k a) `(unbind-function ',k ,a))
(define (bind-function k v a)
  (acons k v (unbind-function k a)))
(define-macro (bind1 k v a) `(acons ',k ,v (unbind ,k ,a)))
(define-macro (bind . bindings-and-a)
  (if (null? (cdr bindings-and-a))
      (car bindings-and-a)
      `(bind1 ,(car bindings-and-a)
              ,(cadr bindings-and-a)
              (bind ,@(cddr bindings-and-a)))))
(define (lookup-function k a)
  (cond ((null? a)
         (error "lookup-function: unbound key" k))
        ((eq? (caar a) k)
         (cdar a))
        (else
         (lookup-function k (cdr a)))))
(define-macro (lookup k a)
  `(lookup-function ',k ,a))
(define (bound-function? k a)
  (cond ((null? a)
         #f)
        ((eq? (caar a) k)
         #t)
        (else
         (bound-function? k (cdr a)))))
(define-macro (bound? k a)
  `(bound-function? ',k ,a))

(define (move object bounce)
  (if (not object)
      #f
      (let* ((x (lookup x object))
             (y (lookup y object))
             (dx (lookup dx object))
             (dy (lookup dy object))
             (min-x (lookup min-x object))
             (min-y (lookup min-y object))
             (max-x (lookup max-x object))
             (max-y (lookup max-y object))
             (unsaturated-x (+ x dx))
             (unsaturated-y (+ y dy))
             (new-dx (if (and bounce (not (within? unsaturated-x min-x max-x)))
                         (- dx)
                         dx))
             (new-dy (if (and bounce (not (within? unsaturated-y min-y max-y)))
                         (- dy)
                         dy)))
        (bind x (saturate unsaturated-x min-x max-x)
              y (saturate unsaturated-y min-y max-y)
              dx new-dx
              dy new-dy
              object))))

(define client-command-lines (cdr (command-line)))

;;; Let's make buffers large enough to carry an entire
;;; long message in the worst case (To do: this can be
;;; computed at startup time)
;(set! PIPE_BUF (* 1024 1024))

(define (run pipes)
  (let loop ((state initial-state)
             (round-no 0))
    ;;(format #t "\n\n-------------------------------------------\nRunning round ~s...\n" round-no)
    (send-state pipes state round-no)
    (let* ((commands (receive-commands pipes))
           (state-with-no-messages (bind messages '() state))
           (state-after-commands (execute-commands commands
                                                   state-with-no-messages
                                                   round-no
                                                   0))
           (state-after-update (update-state state-after-commands
                                             round-no)))
      (loop state-after-update
            (+ round-no 1)))))
    ;; (let* ((state-after-update
    ;;         (update-state state round-no))
    ;;        (state-after-interactions
    ;;         (run-interactions pipes state-after-update round-no 0)))
    ;;   (loop state-after-interactions
    ;;         (+ round-no 1)))))

(define (send-state pipes state round-no)
  (let ((ids (iota (length pipes))))
    (let loop ((pipes pipes)
               (ids ids))
      (if (null? pipes)
          #t
          (begin
            (write (bind your-id (car ids)
                         round round-no
                         state)
                   (car pipes))
            (newline (car pipes))
            (force-output (car pipes))
            (loop (cdr pipes)
                  (cdr ids)))))))

(define (receive-commands pipes)
  (let ((ids (iota (length pipes))))
    (let loop ((pipes pipes)
               (reversed-commands '()))
      (if (null? pipes)
          (reverse reversed-commands)
          (let ((command (read (car pipes))))
            (loop (cdr pipes)
                  (cons command reversed-commands)))))))
            

(define (execute-commands commands state round-no first-id)
  (if (null? commands)
      state
      (let ((state-after-first-command (execute-command (car commands)
                                                        state
                                                        round-no
                                                        first-id)))
        (execute-commands (cdr commands)
                          state-after-first-command
                          round-no
                          (+ first-id 1)))))

(define (randomize-random-state!)
  (let ((time (gettimeofday)))
    (set! *random-state*
          (seed->random-state (+ (car time)
                                 (cdr time)
                                 (getpid))))))
(randomize-random-state!)

(define (average . elements)
  (/ (apply + elements)
     (length elements)))

(define pi/2 (acos 0))
(define pi (* 2 pi/2))

(define (make-ball)
  (let* ((velocity-modulo ball-velocity-modulo)
         (angle (random-from-to 0 (* 2 pi)))
         (dx (* velocity-modulo (cos angle)))
         (dy (* velocity-modulo (sin angle))))
    (bind x      (average left right)
          y      (average top bottom)
          dx     dx
          dy     dy
          min-x  left
          max-x  right
          min-y  top
          max-y  bottom
          width  ball-width
          height ball-height
          '())))

(define (make-balls)
  (map (lambda whatever (make-ball))
       (iota ball-no)))      

(define (make-paddle y)
  (bind x      (random-from-to left right)
        y      y
        dx     0
        dy     0
        min-x  left
        max-x  right
        min-y  y
        max-y  y
        width  paddle-width
        height paddle-height
        '()))

(define initial-state
  `((balls . ,(make-balls))
    (players . ())
    (top-score . 0)
    (bottom-score . 0)
    (messages . ())))

(define (bounce-off-paddle ball paddle)
  (let* ((ball-x (lookup x ball))
         (ball-y (lookup y ball))
         (ball-dx (lookup dx ball)) 
         (ball-dy (lookup dy ball)) 
         (paddle-x (lookup x paddle))
         (paddle-y (lookup y paddle))
         (paddle-dx (lookup dx paddle))
         (paddle-dy (lookup dy paddle))
         (delta-x (abs (- ball-x paddle-x)))
         (delta-y (abs (- ball-y paddle-y))))
    (if (and (<= delta-y (+ ball-radius (/ paddle-height 2)))
             (<= delta-x (/ paddle-width 2)))
        (bind dy (- ball-dy)
              dx (+ ball-dx paddle-dx)
              ball)
        ball)))

(define (bounce-off-paddles ball players)
  (if (null? players)
      ball
      (bounce-off-paddles (bounce-off-paddle ball (car players))
                          (cdr players))))

(define (score-string top-score bottom-score)\
  (string-append "top: "
                 (number->string top-score)
                 " - bottom: "
                 (number->string bottom-score)))

(define (update-state state just-finished-round-no)
  ;;(format #t "\nComputing the next state...\n")
  (let* ((balls (lookup balls state))
         (moved-balls (map (lambda (ball)
                             (move ball #t))
                           balls))
         (moved-players (map (lambda (pair)
                               (cons (car pair)
                                     (move (cdr pair) #f)))
                             (lookup players state)))
         (balls-after-paddle-bounce (map (lambda (moved-ball)
                                           (bounce-off-paddles moved-ball (map cdr moved-players)))
                                         moved-balls))
         (messages (lookup messages state))
;;          (ball-ys (map (lambda (ball-after-paddle-bounce)
;;                          (lookup y ball-after-paddle-bounce))
;;                        balls-after-paddle-bounce))
         (ball-min-y (lookup min-y (car balls-after-paddle-bounce)))
         (ball-max-y (lookup max-y (car balls-after-paddle-bounce)))
         (new-balls '())
         (top-score (lookup top-score state))
         (bottom-score (lookup bottom-score state)))
    (iterate
     (lambda (ball-after-paddle-bounce)
       (let* ((ball-y (lookup y ball-after-paddle-bounce))
              (distance-from-top (- ball-y ball-min-y))
              (distance-from-bottom (- ball-max-y ball-y)))
         (if (or (<= distance-from-top (/ paddle-height 2))
                 (<= distance-from-bottom (/ paddle-height 2)))
             (begin
               (if (< distance-from-top distance-from-bottom)
                   (begin
                     (set! bottom-score (+ bottom-score 1))
                     (set! messages (cons (string-append "The BOTTOM player scored: "
                                                         (score-string top-score bottom-score))
                                          messages)))
                   (begin
                     (set! top-score (+ top-score 1))
                     (set! messages (cons (string-append "The TOP player scored: "
                                                         (score-string top-score bottom-score))
                                          messages))))
               (set! new-balls (cons (make-ball) new-balls)))
             (set! new-balls (cons ball-after-paddle-bounce new-balls)))))
     balls-after-paddle-bounce)
    (bind balls new-balls
          players moved-players
          messages messages
          top-score top-score
          bottom-score bottom-score
          state)))
    
(define (execute-command command state round-no player-id)
  (let* ((players (lookup players state))
         (messages (lookup messages state))
         (current-player-no (length players))
         ;; The first player to join will be on the top
         (y (if (zero? current-player-no) top bottom)))
    (cond ((> current-player-no 2)
           (error "Too many players"))
          ((eq? (car command) 'join)
           (set! messages (cons (string-append "Welcome client n."
                                               (number->string player-id))
                                messages))
           (set! players (bind-function player-id (make-paddle y) players)))
          ((or (eq? (car command) 'left)
               (eq? (car command) 'right))
           (let ((our-player-alist (lookup-function player-id players))
                 (dx (if (eq? (car command) 'left)
                         (- player-dx-modulo)
                         player-dx-modulo)))
             (set! players
                   (bind-function player-id
                                  (bind dx dx our-player-alist)
                                  players))
             #t))
          (else
           #t))
    (bind players players
          messages messages
          state)))

(run (map open-input-output-pipe
          client-command-lines))
