;;; This file is part of guile-with-graphics, a graphic library for Guile

;;; Copyright (C) 2009  Luca Saiu
;;; Updated in 2010 by Luca Saiu

;;; This program is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(use-modules (srfi srfi-9)) ;; this is needed for records

;;; The game field size, in pixels:
(define field-width  800)
(define field-height 600)

;;; The "border" size, i.e. the number of pixels within the field which are
;;; impossible to reach for players:
(define field-border 20)
(define minimum-x-in-the-field field-border)
(define minimum-y-in-the-field field-border)
(define maximum-x-in-the-field (- field-width field-border))
(define maximum-y-in-the-field (- field-height field-border))
(define (in-field-x? x)
  (and (<= minimum-x-in-the-field x)
       (<= x maximum-x-in-the-field)))
(define (in-field-y? y)
  (and (<= minimum-y-in-the-field y)
       (<= y maximum-y-in-the-field)))

;;; Player:
(define player-radius 12) ;12)
(define 2player-radius (* player-radius 2))

;;; How many updates per second to compute and to send to clients (by default).
;;; Notice that about ten decisions per seconds are more enough to play well;
;;; it's still faster than the human reaction time, and with a small frequency
;;; we don't stress the processor too much.
(define default-server-update-recomputation-frequency 62)
(define default-server-update-sending-frequency 30)
(define server-update-sending-frequency-for-client 10);60)
(define server-update-sending-frequency-for-ai-client 15)

;;; We represent the current state of each pill as a record:
(define-record-type pill
  (make-pill-by-specifying-all-fields x y velocity-x velocity-y)
  pill?
  (x                 pill-x                 pill-set-x!)
  (y                 pill-y                 pill-set-y!)
  (velocity-x        pill-velocity-x        pill-set-velocity-x!)
  (velocity-y        pill-velocity-y        pill-set-velocity-y!))

(define (make-pill)
  (make-pill-by-specifying-all-fields (random-x-in-the-field)
                                      (random-y-in-the-field)
                                      (random-from-to (- maximum-pill-component-velocity-in-modulo)
                                                      maximum-pill-component-velocity-in-modulo)
                                      (random-from-to (- maximum-pill-component-velocity-in-modulo)
                                                      maximum-pill-component-velocity-in-modulo)))

;;; Pills:
(define pills-no 20);30)
(define pill-radius 5)
(define pill-eating-radius (max (* player-radius 3/2)
                                (* pill-radius 3/2)))

;;; Maximum translation velocity in modulo for pills, for each component (x and y),
;;; in pixels per second:
(define maximum-pill-component-velocity-in-modulo
  50)

;;; Rotation velocity, in radiants per second:
(define rotation-velocity 9)

;;; Translation velocity, in pixels per second:
(define translation-velocity 200);300)

;;; Compute random reachable coordinates within the field:
(define (random-x-in-the-field)
  (random-from-to field-border maximum-x-in-the-field))
(define (random-y-in-the-field)
  (random-from-to field-border maximum-y-in-the-field))

;;; Saturate coordinates (meaning: return them unchanged if they are within
;;; the game field, otherwise return the appropriate border value):
(define (saturate-x x)
  (saturate x field-border maximum-x-in-the-field))
(define (saturate-y y)
  (saturate y field-border maximum-y-in-the-field))

;;; Compute the Euclidean distance from the given position to the given pill:
(define (distance-from-pill x y pill)
  (distance (pill-x pill) (pill-y pill) x y))

;;; Is the given pill within the eating range from the given position?
(define (eaten? x y pill)
  ;; Notice that a pill is represented as the cons of its coordinates:
  (<= (distance-from-pill x y pill)
      pill-eating-radius))

;;; Return the angle which must be added to the given angle of an object in the
;;; given position so that the object in the given position is oriented towards
;;; the given point:
(define (angle-to-target x y angle target-x target-y)
  (let* ((delta-x (- target-x x))
         (delta-y (- y target-y))
         (delta-x (if (zero? delta-x) 0.0001 delta-x)))
    (normalize-angle (- (if (> delta-x 0)
                            (atan (/ delta-y delta-x))
                            (+ pi (atan (/ delta-y delta-x))))
                        angle))))
