;;; This file is part of guile-with-graphics, a graphic library for Guile

;;; Copyright (C) 2009  Luca Saiu

;;; This program is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;;; The user is only interested in calling console-write-line!, console-clear!
;;; and (once per tick) console-draw.  Our neat scrolling effects are
;;; automatically supported.
;;; The user can also change console-foreground-color and
;;; console-background-color with a simple set!.


;;; The user should *not* load this file directly. It's already loaded by
;;; interactive.scm.

(define console-external-border 10)
(define console-internal-border 5)
(define console-line-no 5)
(define console-inter-line-spacing 2)

;;; If the background color is #f instead of a color, then the console is drawn
;;; with no background. This variable can be safely changed with set! after
;;; initialization:
(define console-background-color #f)
;(define console-background-color (make-color 0.2 0.0 0.0))

(define console-foreground-color (make-color 1 1 1))
;;(define console-foreground-input-color (light yellow))

;;; The console state can be one of of hidden, shown, showing or hiding.
(define console-state
  'hidden)

;;; A ratio of 1 means that the console is fully shown, a ratio of 0 means that
;;; it is fully hidden. All values in [0, 1] are valid.
(define console-show-ratio 0)

;;; The absolute time (as returned by milliseconds-since-startup) when the
;;; state changed the last time:
(define console-last-state-change-time 0)

(define console-full-showing-time-after-each-change 3000) ; in milliseconds

;;; How long the console takes, in milliseconds, for showing up or hiding:
(define console-showing-time 250);;500)
(define console-hiding-time 2000)

;;; The list of lines currently shown in the console, from the oldest to the newest,
;;; as a list of strings
(define console-lines
  '())
;;'("abc" "123" "foo bar baaz quux" "boo" "this is the bottom line"))

(define (console-update-state!)
  (let ((current-time (milliseconds-since-startup)))
    (case console-state
      ((hidden)
       ;; Do nothing
       #t)
      ((shown)
       (if (>= current-time (+ console-last-state-change-time
                               console-full-showing-time-after-each-change))
           (begin
             (set! console-last-state-change-time current-time)
             (set! console-state 'hiding))))
      ((showing)
       (if (>= current-time (+ console-last-state-change-time
                               console-showing-time))
           (begin
             (set! console-last-state-change-time current-time)
             (set! console-state 'shown)
             (set! console-show-ratio 1))
           (set! console-show-ratio
                 (/ (- current-time console-last-state-change-time)
                    console-showing-time))))
      ((hiding)
       (if (>= current-time (+ console-last-state-change-time
                               console-hiding-time))
           (begin
             (set! console-last-state-change-time current-time)
             (set! console-state 'hidden)
             (set! console-show-ratio 0))
           (set! console-show-ratio
                 (- 1
                    (/ (- current-time console-last-state-change-time)
                       console-hiding-time)))))
      (else
       (error "console-update-state!: Unknown console state " console-state)))))

(define (console-start-showing!)
  (let ((current-time (milliseconds-since-startup)))
    (case console-state
      ((shown)
       ;; Make the current state last longer:
       (set! console-last-state-change-time current-time))
      ((showing)
       ;; Do nothing, the console is already showing:
       #t)
      ((hiding hidden)
       ;; Don't change the show ration, but change the scroll direction; we have to
       ;; "lie" about console-last-state-change-time:
       (set! console-last-state-change-time
             (- current-time
                (* console-showing-time console-show-ratio)))
       (set! console-state 'showing))
      (else
       (error "console-stard-showing!: Unknown console state " console-state)))))

(define (console-draw)
  ;; Compute the current console state, so that it's in the correct position when
  ;; we draw it:
  (console-update-state!)
  (unless (zero? console-show-ratio) ;; do nothing if the console is completely hidden
    (let* ((console-x1 console-external-border)
           (console-width (- (window-width) (* 2 console-external-border)))
           (console-height (+ (* console-line-no (font-height))
                              (* console-line-no console-inter-line-spacing)
                              (* 2 console-internal-border)))
           (console-y1 (- (window-height)
                          (* console-show-ratio
                             (+ console-external-border
                                console-height))))
           (first-line-x1 (+ console-x1 console-internal-border))
           (first-line-y1 (+ console-y1 console-internal-border)))
      (if console-background-color
          (draw-rectangle console-x1 console-y1 console-width console-height console-background-color))
      (console-draw-text console-lines first-line-x1 first-line-y1))))
  
(define (console-draw-text console-lines first-line-x1 first-line-y1)
  (if (null? console-lines)
      'done
      (begin
        (draw-text (car console-lines) first-line-x1 first-line-y1 console-foreground-color)
        (console-draw-text (cdr console-lines) first-line-x1 (+ first-line-y1
                                                                console-inter-line-spacing
                                                                (font-height))))))

(define (console-write-line! new-line)
  (console-start-showing!)
  (set! console-lines
        (append (if (= (length console-lines) console-line-no)
                    (cdr console-lines)
                    console-lines)
                (list new-line))))

(define (console-clear!)
  (set! console-lines '()))
